#include "lineSetToQuadMesh.h"
#include <math.h>
#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoLineSet.h>
#include <Inventor/nodes/SoQuadMesh.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoTransform.h>

void
endCap(SoSeparator *root, SbVec3f &position, float radius)
  {
  SoTransform *transform = new SoTransform();
  transform->translation.setValue(position);
  root->addChild(transform);

  static SoSphere    *capSphere = new SoSphere();
  capSphere->radius = radius;
  root->addChild(capSphere);
  }

////////////////////////////////////////////////////////////
//
//                   ^ u3
//                   |    
//                *  |  * 
//                   |     
//            *      |      * 
//                   |       
//                   |       
//         --*---------------*---> u2
//                   |
//                   |
//            *      |      * 
//                   |
//                *  |  * 
//                   |
//
//


////////////////////////////////////////////////////////////
//
//                       root
//                        |
//          ----------------------- ..  -----
//          |              |                |       
//      tubeRoot        tubeRoot        tubeRoot
//

void
lineSetToQuadMesh(SoSeparator *root, SoCoordinate3 *lineVerts,
		  SoLineSet *lineSet, float radius, int numSamples)
  {
  int lineNum, numLines;
  int vertNum; 
  const int32_t *numVerts;
  int index;
  int sampleNum;
  float theta, dTheta;
  SbVec3f u1, u2, u3;           // local coordinate system
  const SbVec3f *lineVert; 
  SbVec3f center;

  root->removeAllChildren();


  // Loop through the polylines in the lineSet.
  // 
  // lineSet->numVertices is a multi-valued field; it holds an array.
  // 
  // lineSet->numVertices->getNum() tells how many elements are in the array.
  //
  // lineSet->numVertices->getValues(k) array-of-sizes starting at index k.
  //


  numLines = lineSet->numVertices.getNum();
  numVerts = lineSet->numVertices.getValues(0);

  lineVert = lineVerts->point.getValues(0);  // initial element in vertex-array

  index = 0;  // index will increment through the points in the array.

  dTheta = 2.0*M_PI/numSamples;

  SoSeparator *tubeRoot;
  SoCoordinate3 *tubeMesh;
  SbVec3f *tubeVerts;
  SoQuadMesh *quadMesh; 

  ////////////////////////////////////////////////////////////
  //
  //                    tubeRoot
  //                        |
  //                ----------------
  //                |              |
  //            tubeMesh        quadMesh
  //

  if (numLines == 0) return;

  for (lineNum = 0; lineNum < numLines; lineNum++)
    {
    if (numVerts[lineNum] < 2) continue;

    tubeRoot = new SoSeparator();
    root->addChild(tubeRoot);

    tubeMesh = new SoCoordinate3();
    tubeMesh->point.setNum(numVerts[lineNum] * (numSamples+1));
    tubeRoot->addChild(tubeMesh);

    quadMesh = new SoQuadMesh();
    tubeRoot->addChild(quadMesh);

    tubeVerts = tubeMesh->point.startEditing();

    u2 = SbVec3f(1.0, 0.0, 0.0);   // tentative choice of u2
    u3 = SbVec3f(0.0, 0.0, 1.0);   // tentative choice of u3

    for (vertNum = 0; vertNum < numVerts[lineNum]; vertNum++, index++)
      {
      center = lineVert[index];

      if (vertNum < numVerts[lineNum] - 1)
	{
	u1 = lineVert[index+1] - center;
	}
      else
	{
	u1 = center - lineVert[index-1];
	}
      
      u1.normalize();
      u2.normalize();
      u2 = u2 - (u1.dot(u2))*u1;
      if (fabs(u2.dot(u2)) <= 0.001)
	{
	u2 = u3;
	u2.normalize();
	u2 = u2 - (u1.dot(u2))*u1;
	}
      u2.normalize();

      u3 = u1.cross(u2);

      for (theta = 0.0, sampleNum = 0; sampleNum <= numSamples; sampleNum++, theta += dTheta)
	{
	if (sampleNum == numSamples) theta = 0.0;

	tubeVerts[vertNum*(numSamples + 1) + sampleNum] =  center + 
	  radius*cos(theta)*u2 + radius*sin(theta)*u3;
	}
      }
    tubeMesh->point.finishEditing();

    quadMesh->verticesPerRow = numSamples + 1;
    quadMesh->verticesPerColumn = numVerts[lineNum];
    }

  }

